/*
    ansiman.c - displays mandelbrot set with ANSI.SYS colors
    Copyright (C) 1994, 1995 Jeff Robertson

    version 1.2, May 14, 1995

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    The author of this program can be contacted at one of the
    following e-mail addresses:

    jrobert1@sss.cba.ua.edu
    jrobert1@midgard.cba.ua.edu
    jrobert1@job.cba.ua.edu
    jroberts@buster.eng.ua.edu
    jrobert1@ua1ix.ua.edu
    jrobert1@ua1vm.ua.edu

    Or snail-mail:

    P.O. Box CO
    Tuscaloosa, AL 35487

*/

#include <stdio.h>
#include <float.h>
#include <stdlib.h>

#define BAILOUT 2
#define MAXITER 200

#define SCREEN_ROWS 23
#define SCREEN_COLS 78
#define PALLETTE 6  /* number of colors to use */
#define INSIDE ' '
#define SHADE 176
#define SOLID 219

#define VERSION "1.2"

/* in version 1, this actually set color to white on black */
#define reset_color fputs("\x1b[0m\x1b[255D", stdout)

struct complex { double x,y;};

double get_float( double *, char *, int);
void work_horse( double, double, double, double);
void copying(void);
void warranty(void);
void version(void);
int mandel( struct complex );


int shade=1, verbose=1, one_shot=0;

main( int argc, char *argv[])
{
    double x_max, x_min, y_max, y_min;
    int k;

    /* get options */
    for(k=1; k < argc ; k++)
    {
        /* -f turns off prompting */
        if (!strcmp(argv[k], "-f"))
            verbose = 0;

        /* -s turns off shading */
        else if (!strcmp(argv[k], "-s"))
            shade = 0;

        /* -c get coords */
        else if ( !strcmp(argv[k], "-c") )
        {
            one_shot = 1;
            x_min = atof( argv[++k]);
            y_max = atof( argv[++k]);
            x_max = atof( argv[++k]);
            y_min = atof( argv[++k]);
        }

        /* version information */
        else if ( !strcmp(argv[k], "-v"))
        {
            version();
            return 0;
        }                        

        /* help! */
        else if ( !strcmp(argv[k],"-h"))
        {
            version();
            puts("\noptions:");
            puts("-v    version information");
            puts("-i    copyright and warranty information");
            puts("-f    turn off prompting and messages");
            puts("-s    turn off shading");
            puts("-c    (followed by coords) use coords given");
            return 0;
         }

        else if ( !strcmp(argv[k],"-i"))
        {
            version();
            putchar('\n');
            copying();
            warranty();
            return 0;
        }
    }

#ifdef debug
    printf("verbose: %d, shade: %d, one_shot: %d\n",verbose,shade,one_shot);
#endif


    if (one_shot)
    {
        work_horse(x_min, y_max, x_max, y_min);
        return 0;
    }
    
    while( 'A' == 'A')
    {
        if (verbose)
        {
            reset_color;
            puts("\n");
            version();

            printf("ANSIMAN has ABSOLUTELY NO WARRANTY;\n");
            printf("Enter 'w' for details.\n");
            printf("This is free software, and you are welcome\n");
            printf("to redistribute it under certain conditions.\n");
            printf("Enter 'c' for details.\n");
            printf("Enter 'q' or end-of-file to exit.\n\n");

            printf("Define viewing window: ( -2, 1, 1, -1 is good)\n");
        }
        get_float(&x_min,"X coordinate of upper left corner: ", verbose);
        get_float(&y_max,"Y coordinate of upper left corner: ", verbose);
        get_float(&x_max,"X coordinate of lower right corner: ", verbose);
        get_float(&y_min,"Y coordinate of lower right corner: ", verbose);

        if (verbose)
            printf("\n\nOkay, here it is....\n");
        work_horse( x_min, y_max, x_max, y_min);
        if (verbose)
        {
            printf("-press <enter> to go back to menu-");
            getchar();
        }
    }

    return(0);
}

/* color the points, one by one */
void work_horse(double x_min, double y_max, double x_max, double y_min)
{
    struct complex z;   /* used over and over for each point */
    double x_res, y_res; /* distance from one point to the next */

    /* these declarations used to be inside the for() loops */
    int i, color, last_color;

    x_res = (( x_max - x_min) / SCREEN_COLS );
    y_res = (( y_max - y_min) / SCREEN_ROWS );

    for( z.y = y_max; z.y >= y_min ; z.y-= y_res)
    {
        last_color = -1;
        for( z.x = x_min; z.x<= x_max; z.x+= x_res)
        {
            i = mandel(z);

            /* this optimizes the ANSI by only using color */
            /* codes when they are actually needed.        */
            color = i% ( shade ? (PALLETTE * PALLETTE) : PALLETTE );
            if ( i != MAXITER)
            {
                if (color != last_color)
                {
                    if (shade)
                    {
                        /* foreground */
                        fputs( "\x1b[1;3", stdout);
                        putchar( (color % PALLETTE) + '1');
                        fputs(";4", stdout);

                        /* background */
                        putchar( (color / PALLETTE) + '1');
                        putchar('m');
                    }
                    else
                    {
                        fputs("\x1b[1;3",stdout);
                        putchar(color + '1');
                        putchar('m');
                    }
                    last_color = color;
                }
                putchar( shade ? SHADE : SOLID );
            }
            else
            {
                if (last_color != -1)
                {
                    last_color = -1;
                    fputs("\x1b[40m",stdout);
                }
                putchar(INSIDE);
            }
        }
        puts("\x1b[40m");
    }
    reset_color;
}

/* prompt for and read floating point number from stdin, or exit on EOF */
double get_float(double *num, char *message, int verbose)
{
    void copying(void);
    void warranty(void);
    char foo[16];

    while(1)
    {
        if (verbose)
            printf("%s", message);
        if (!fgets(foo, sizeof(foo), stdin) || !strcmp(foo,"q\n"))
            exit(0);
        else if (!strcmp(foo, "c\n"))
            copying();
        else if (!strcmp(foo, "w\n"))
            warranty();
        else
            break;
    }
    sscanf(foo,"%lf", num);
    return *num;
}

/* return the number of iterations of a complex point   */
/* before escaping from the Mandelbrot set, if ever.    */
/* return MAXITER if it is reached.                     */
int mandel( struct complex z)
{

    int got_away = 0, iter;
    struct complex c, temp;
    c = z;

    for (iter=0; (iter<MAXITER)&&(!got_away) ; iter++)
    {
        /* z = z^2 + c */
        temp = z;
        z.x = (temp.x * temp.x) -  (temp.y * temp.y);
        z.y = 2 * temp.x * temp.y;
        z.x += c.x;
        z.y += c.y;

        got_away = ((z.x*z.x + z.y*z.y) >= (BAILOUT*BAILOUT));
    }
    return iter;
}

void version(void)
{
    puts("ANSI Mandelbrot Set");
    printf("version %s\n", VERSION);
    puts("(c) 1994 Jeff Robertson");
}

void warranty(void)
{
    puts("This program is distributed in the hope that it will be useful,");
    puts("but WITHOUT ANY WARRANTY; without even the implied warranty of");
    puts("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the");
    puts("GNU General Public License for more details.");
}

void copying(void)
{
    puts("This program is free software; you can redistribute it and/or modify");
    puts("it under the terms of the GNU General Public License as published by");
    puts("the Free Software Foundation; either version 2 of the License, or");
    puts("(at your option) any later version.");
}
